﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using AKS.BusinessModel.Generator;

namespace AKS.BusinessModel.Inventory
{
  /// <summary>
  /// Projekt
  /// </summary>
  public class Project : Persistence.IPersistentObject
  {
    #region Public Enums

    /// <summary>
    /// Projektstatus
    /// </summary>
    public enum State
    {
      /// <summary>
      /// In Vorbereitung
      /// </summary>
      Projected = 1,

      /// <summary>
      /// In Bearbeitung
      /// </summary>
      InProcess = 2,

      /// <summary>
      /// Beendet
      /// </summary>
      Finished = 3,

      /// <summary>
      /// Abgeschlossen
      /// </summary>
      Closed = 4,

      /// <summary>
      /// Abgebrochen
      /// </summary>
      Canceled = 5
    }

    #endregion

    #region Private Members

    /// <summary>
    /// Daten
    /// </summary>
    private Guid id;
    private string name;
    private string owner;
    private string manager;
    private DateTime? startDate;
    private DateTime? endDate;

    private State state;
    private Guid? extId;

    // Anlageinstanzen dieses Projektes
    private Dictionary<Guid, FacilityEntityProjectRelation> facilityEntities = new Dictionary<Guid, FacilityEntityProjectRelation>();

    // Orte dieses Projekts
    private Dictionary<Guid, PlaceProjectRelation> places = new Dictionary<Guid, PlaceProjectRelation>();

    /// <summary>
    /// Persistenzzustand dieser Instanz
    /// </summary>
    protected Persistence.PersistentObjectState<Guid, Project> objectState;

    #endregion

    #region Private Undo-Buffers

    /// <summary>
    /// Daten für die Implementierung der IRevertibleChangeTracking- und IEditableObject-Schnittstellen
    /// </summary>
    protected Project origin = null;
    protected Project editOrigin = null;

    #endregion

    #region Public Constructors

    /// <summary>
    /// Konstruktor zum Erstellen eines neuen Projektes
    /// </summary>
    /// <param name="pm">Persistenzmanager</param>
    public Project( Persistence.PersistenceManager pm )
    {
      this.objectState = new Persistence.PersistentObjectState<Guid, Project>( pm != null ? pm.ProjectState : null );
      this.id = Guid.NewGuid();
      this.name = this.owner = this.manager = string.Empty;
      this.startDate = this.endDate = null;

      this.state = State.Projected;
      this.extId = null;

      this.objectState.Created( this.id, this );
    }

    /// <summary>
    /// Konstruktor zum Erstellen eines vorhandenen Projektes
    /// </summary>
    /// <param name="pm">Persistenzmanager</param>
    /// <param name="id">ID des Projekts</param>
    /// <param name="name">Projektname</param>
    /// <param name="owner">Projekteigentümer</param>
    /// <param name="manager">Projektleiter</param>
    /// <param name="startDate">Startdatum des Projekts</param>
    /// <param name="endDate">Enddatum des Projekts</param>
    /// <param name="stateCode">Status-Code - dies ist der Int-Wert des Typ-Enums</param>
    /// <param name="extId">ID des Esxternen Projektes oder null, für interne Projekte</param>
    public Project( Persistence.PersistenceManager pm, 
                    Guid id, string name, string owner, string manager, DateTime? startDate, DateTime? endDate,
                    int stateCode, Guid? extId )
    {
      this.objectState = new Persistence.PersistentObjectState<Guid, Project>( pm != null ? pm.ProjectState : null );
      this.id = id;
      this.name = name;
      this.owner = owner;
      this.manager = manager;
      this.startDate = startDate;
      this.endDate = endDate;
      this.state = (State)stateCode;
      this.extId = extId;
    }

    #endregion

    #region Public Events

    /// <summary>
    /// Event für die NotifyPropertyChanged-Ereignisse
    /// </summary>
    public event System.ComponentModel.PropertyChangedEventHandler PropertyChanged;

    #endregion

    #region Public Properties

    /// <summary>
    /// ID des Projekts
    /// </summary>
    public Guid Id
    {
      get { return this.id; }
    }

    /// <summary>
    /// Projektname
    /// </summary>
    public string Name
    {
      get { return this.name; }
      set
      {
        if ( this.name == value )
          return;

        this.name = value;
        this.NotifyPropertyChanged( "Name" );
      }
    }

    /// <summary>
    /// Projekteigentümer
    /// </summary>
    public string Owner
    {
      get { return this.owner; }
      set
      {
        if ( this.owner == value )
          return;

        this.owner = value;
        this.NotifyPropertyChanged( "Owner" );
      }
    }

    /// <summary>
    /// Projektleiter
    /// </summary>
    public string Manager
    {
      get { return this.manager; }
      set
      {
        if ( this.manager == value )
          return;

        this.manager = value;
        this.NotifyPropertyChanged( "Manager" );
      }
    }

    /// <summary>
    /// Startdatum des Projekts
    /// </summary>
    public DateTime? StartDate
    {
      get { return this.startDate; }
      set
      {
        if ( this.startDate.GetValueOrDefault( DateTime.MinValue ) == value.GetValueOrDefault( DateTime.MinValue ) )
          return;

        this.startDate = value;
        this.NotifyPropertyChanged( "StartDate" );
      }
    }

    /// <summary>
    /// Enddatum des Projekts
    /// </summary>
    public DateTime? EndDate
    {
      get { return this.endDate; }
      set
      {
        if ( this.endDate.GetValueOrDefault( DateTime.MinValue ) == value.GetValueOrDefault( DateTime.MinValue ) )
          return;

        this.endDate = value;
        this.NotifyPropertyChanged( "EndDate" );
      }
    }

    /// <summary>
    /// Projekstatus
    /// <remarks>Um 'überspringende' Statusübergange zu erzwingen muss die Methods <see cref="ForceStateChange"/> verwendet werden</remarks>
    /// </summary>
    public State ProjectState
    {
      get { return this.state; }
      set
      {
        if ( this.state == value || !this.StateTransitions.Contains( value ) )
          return;

        this.state = value;
        this.NotifyPropertyChanged( "ProjectState" );
        this.NotifyPropertyChanged( "ProjectStateCode" );
      }
    }

    /// <summary>
    /// Typcode des Projektstaus
    /// </summary>
    public int ProjectStateCode
    {
      get { return Convert.ToInt32( this.state ); }
      set { this.ProjectState = (State)value; }
    }

    /// <summary>
    /// ID des externen Projekts
    /// </summary>
    public Guid? ExtId
    {
      get { return this.extId; }
      set
      {
        if ( this.extId == value )
          return;

        this.extId = value;
        this.NotifyPropertyChanged( "ExtId" );
      }
    }

    /// <summary>
    /// Das Projekt ist schreibgeschützt, wenn es eine Live-ID hat
    /// </summary>
    public bool ReadOnly
    {
      get { return this.extId.HasValue; }
    }
    /// <summary>
    /// Dem Projekt zugeordnete Anlageinstanzen
    /// </summary>
    public IList<FacilityEntity> FacilityEntities
    {
      get {
        List<FacilityEntity> facilityEntities = new List<FacilityEntity>();
        foreach ( FacilityEntityProjectRelation rel in this.facilityEntities.Values )
          facilityEntities.Add( rel.FacilityEntity );

        return facilityEntities; 
      }
    }

    /// <summary>
    /// Projekt-Anlageinstanz-Beziehungen
    /// </summary>
    public IList<FacilityEntityProjectRelation> FacilityEntityProjectRelation
    {
      get { return this.facilityEntities.Values.ToList<FacilityEntityProjectRelation>(); }
    }

    /// <summary>
    /// Dem Projekt zugeordnete Orte
    /// </summary>
    public IList<PlaceBase> Places
    {
      get { return this.places.Values.Select( r => r.Place ).ToList(); }
    }

    /// <summary>
    /// Projekt-Ort-Beziehungen
    /// </summary>
    public IList<PlaceProjectRelation> PlaceProjectRelation
    {
      get { return this.places.Values.ToList(); }
    }

    /// <summary>
    /// Liefert die erlaubten Statusübergange
    /// </summary>
    public State[] StateTransitions
    {
      get
      {
        if ( this.state == State.Projected && !this.ExtId.HasValue )
          return new State[] { State.Canceled, State.InProcess };

        if ( this.state == State.InProcess )
          return new State[] { State.Canceled, State.Finished };

        if ( this.state == State.Finished )
          return !this.ExtId.HasValue ? new State[] { State.Canceled, State.InProcess, State.Closed } : new State[] { State.Canceled, State.InProcess };

        if ( this.state == State.Canceled && !this.ExtId.HasValue )
          return new State[] { State.Projected };

        return new State[0];
      }
    }

    #endregion

    #region Public Methods

    /// <summary>
    /// Erstellt eine Kopie des Projektes. Die Kopie verfügt hat keinen Persistenzmanager
    /// </summary>
    /// <returns></returns>
    public object Clone()
    {
      return new Project( null, this.id, this.name, this.owner, this.manager, this.startDate, this.endDate, 
                          this.ProjectStateCode, this.extId );
    }


    /// <summary>
    /// Bestimmt, ob das angegebene Object und das aktuelle Object gleich sind
    /// </summary>
    /// <param name="rhv">Das Object, das mit dem aktuellen Object verglichen werden soll.</param>
    /// <returns>true, wenn das angegebene Object gleich dem aktuellen Object ist, andernfalls false</returns>
    public override bool Equals( object rhv )
    {
      Project project = rhv as Project;
      if ( project == null )
        return false;

      return object.ReferenceEquals( this, rhv ) || this.id == project.id;
    }

    /// <summary>
    /// Hashfunktion
    /// </summary>
    /// <returns>Hashcode für das aktuelle Objekt</returns>
    public override int GetHashCode()
    {
      return id.GetHashCode();
    }

    /// <summary>
    /// Weist dem Projekt die Daten des Arguments zu
    /// </summary>
    /// <param name="rhv">Projekt dessen Daten übernommen werden</param>
    /// <returns>Dieses Projekt</returns>
    public Project Assign( Project rhv )
    {
      this.name = rhv.name;
      this.owner = rhv.owner;
      this.manager = rhv.manager;
      this.startDate = rhv.startDate;
      this.endDate = rhv.endDate;
      this.state = rhv.state;

      this.state = rhv.state;
      this.extId = rhv.extId;

      this.NotifyPropertyChanged( string.Empty );
      return this;
    }

    /// <summary>
    /// Fügt diesem Projekt eine Anlageinstanz hinzu
    /// </summary>
    /// <param name="facilityEntityProjectRelation">Anlageinstanz-Projekt-Beziehung</param>
    public void Add( FacilityEntityProjectRelation facilityEntityProjectRelation )
    {
      if ( this.facilityEntities.ContainsKey( facilityEntityProjectRelation.Id ) )
        return;

      this.facilityEntities.Add( facilityEntityProjectRelation.Id, facilityEntityProjectRelation );
      facilityEntityProjectRelation.FacilityEntity.Add( facilityEntityProjectRelation.Project );
    }

    /// <summary>
    /// Löscht die Beziehung zu einer Anlageinstanz
    /// </summary>
    /// <param name="entity">Anlageinstanz deren Beziehung gelöscht werden soll.</param>
    public void DeleteRelation( FacilityEntity entity )
    {
      List<FacilityEntityProjectRelation> relations = new List<FacilityEntityProjectRelation>();

      // Alle Beziehungen zu der Anlageinstanz
      foreach ( FacilityEntityProjectRelation relation in this.facilityEntities.Values ) {

        if ( relation.FacilityEntity == entity )
          relations.Add( relation );

      }

      // Beziehungen löschen und aus der Liste entfernen
      foreach ( FacilityEntityProjectRelation relation in relations ) {
        relation.Delete();
        this.facilityEntities.Remove( relation.Id );
      }
    }

    /// <summary>
    /// Entfernen einer Anlageinstanz-Beziehung
    /// </summary>
    /// <param name="relation">Anlageinstanz-Beziehung die entfernt wird</param>
    /// <returns>true, wenn die Beziehung entfernt wurde</returns>
    public bool Remove( FacilityEntityProjectRelation relation )
    {
      return this.facilityEntities.Remove( relation.Id );
    }

    /// <summary>
    /// Suchen einer Anlageinstanz-Projekt-Beziehung
    /// </summary>
    /// <param name="id">ID der Anlageinstanz-Projekt-Beziehung</param>
    /// <returns>Anlageinstanz-Projekt-Beziehung oder null, wenn nciht gefunden</returns>
    public FacilityEntityProjectRelation FindFacilityRelation( Guid id )
    {
      return this.facilityEntities.ContainsKey( id ) ? this.facilityEntities[id] : null;
    }

    /// <summary>
    /// Fügt diesem Projekt einen Ort hinzu
    /// </summary>
    /// <param name="placeProjectRelation">Ort-Projekt-Beziehung</param>
    public void Add( PlaceProjectRelation placeProjectRelation )
    {
      if ( this.places.ContainsKey( placeProjectRelation.Id ) )
        return;

      this.places.Add( placeProjectRelation.Id, placeProjectRelation );
      placeProjectRelation.Place.Add( placeProjectRelation.Project );
    }

    /// <summary>
    /// Löscht die Beziehung zu einem Ort
    /// </summary>
    /// <param name="place">Ort dessen Beziehung gelöscht werden soll.</param>
    public void DeleteRelation( PlaceBase place )
    {
      // Alle Beziehungen zum Ort
      List<PlaceProjectRelation> relations = new List<PlaceProjectRelation>( this.places.Values.Where( r => r.Place == place ) );

      // Beziehungen löschen und aus der Liste entfernen
      foreach ( PlaceProjectRelation relation in relations ) {
        relation.Delete();
        this.places.Remove( relation.Id );
      }
    }

    /// <summary>
    /// Entfernen einer Ort-Projekt-Beziehung
    /// </summary>
    /// <param name="relation">Ort-Projekt-Beziehung die entfernt wird</param>
    /// <returns>true, wenn die Beziehung entfernt wurde</returns>
    public bool Remove( PlaceProjectRelation relation )
    {
      return this.places.Remove( relation.Id );
    }

    /// <summary>
    /// Suchen einer Ort-Projekt-Beziehung
    /// </summary>
    /// <param name="id">ID der Ort-Projekt-Beziehung</param>
    /// <returns>Ort-Projekt-Beziehung oder null, wenn nicht gefunden</returns>
    public PlaceProjectRelation FindPlaceRelation( Guid id )
    {
      return this.places.ContainsKey( id ) ? this.places[id] : null;
    }

    #endregion

    #region Protected Methods

    /// <summary>
    /// Implementierung des Notify-Patterns für die Datenbindung
    /// </summary>
    /// <param name="propertyName">Name der Eigenschaft die geändert wurde. Null für alle Eigenschaften</param>
    protected void NotifyPropertyChanged( string propertyName )
    {
      this.objectState.Changing( this.id, this );
      OnPropertyChanged( this, new PropertyChangedEventArgs( propertyName ) );
    }

    /// <summary>
    /// Notify-Event auslösen
    /// </summary>
    protected virtual void OnPropertyChanged( Object sender, System.ComponentModel.PropertyChangedEventArgs e )
    {
      if ( PropertyChanged != null )
        PropertyChanged( sender, e );
    }

    /// <summary>
    /// Macht eine Änderung wieder rückgängig
    /// </summary>
    /// <param name="origin">Projekt mit den Originalwerten</param>
    protected void Undo( Project origin )
    {
      if ( origin == null )
        return;

      this.name = origin.name;
      this.owner = origin.owner;
      this.manager = origin.manager;
      this.startDate = origin.startDate;
      this.endDate = origin.endDate;
      this.state = origin.state;
      this.extId = origin.extId;

      this.NotifyPropertyChanged( string.Empty );
    }

    #endregion

    #region IPersistentObject Member

    /// <summary>
    /// Objektzustand
    /// </summary>
    public Persistence.ObjectState ObjectState
    {
      get { return this.objectState.State; }
    }

    /// <summary>
    /// Persistenzmanager
    /// </summary>
    public Persistence.PersistenceManager PersistenceManager
    {
      get { return this.objectState.PersistentState.PersistenceManager; }
    }

    /// <summary>
    /// Markiert das Objekt als geändert 
    /// </summary>
    public void Changed()
    {
      this.objectState.Changed( this.id, this );
    }

    /// <summary>
    /// Markiert das Objekt als gelöscht 
    /// </summary>
    public void Delete()
    {
      foreach ( PlaceProjectRelation relation in this.places.Values )
        relation.Delete();

      foreach ( FacilityEntityProjectRelation relation in this.facilityEntities.Values )
       relation.Delete();

      this.objectState.Deleted( this.Id, this );
    }

    #endregion

    #region IEditableObject Member

    /// <summary>
    /// Beginnt die Bearbeitung des Ortes. Der Originalzustand wird gespeichert.
    /// </summary>
    void IEditableObject.BeginEdit()
    {
      this.editOrigin = this.Clone() as Project;
    }

    /// <summary>
    /// Verwirft die Änderungen seit dem letzten Aufruf von BeginEdit
    /// </summary>
    void IEditableObject.CancelEdit()
    {
      if ( this.editOrigin != null )
        this.Undo( this.editOrigin );

      this.editOrigin = null;
    }

    /// <summary>
    /// Beendet die Bearbeitung des Ortes. 
    /// </summary>
    void IEditableObject.EndEdit()
    {
      // falls noch keine Originalzustand vorhanden ist, diesen übernehmen
      if ( this.origin == null && this.IsChanged )
        this.origin = this.editOrigin;

      this.Changed();
      this.editOrigin = null;
    }

    #endregion

    #region IRevertibleChangeTracking Member

    /// <summary>
    /// Setzt den Zustand auf seine Originalwerte zurück.
    /// </summary>
    public void RejectChanges()
    {
      this.Undo( this.origin );
      this.origin = null;

      this.objectState.RejectChanges( this.id, this );
    }

    /// <summary>
    /// Akzeptiert die Änderungen
    /// </summary>
    public void AcceptChanges()
    {
      this.origin = null;

      this.objectState.AcceptChanges( this.id, this );
    }

    /// <summary>
    /// Dieses Projekt wurde geändert
    /// </summary>
    public bool IsChanged
    {
      get { return this.objectState.State != Persistence.ObjectState.Unchanged; }
    }

    /// <summary>
    /// Forciert den Statuswechsel, auch wenn er in der Rehienfolge nicht zulässig ist
    /// </summary>
    /// <param name="state">neuer Projektstatus</param>
    public void ForceStateChange( State state )
    {
      if ( this.state == state )
        return;

      this.state = state;
      this.NotifyPropertyChanged( "ProjectState" );
      this.NotifyPropertyChanged( "ProjectStateCode" );
    }

    #endregion

  }

  /// <summary>
  /// Container für Projekte als Dictionary
  /// </summary>
  public class Projects : Dictionary<Guid, Project>
  {

    #region Public Constructors

    /// <summary>
    /// Standardkonstruktor
    /// </summary>
    public Projects()
      : base()
    {
    }

    /// <summary>
    /// Konstruktor der diese Instanz mit den übergebenen Projekten initialisiert
    /// </summary>
    /// <param name="projects">Projekte</param>
    public Projects( IDictionary<Guid, Project> projects )
      : base( projects )
    {
    }

    #endregion

    #region Public Properties

    /// <summary>
    /// Liste der internen Projekte
    /// </summary>
    public IList<Project> InternalProjects
    {
      get { return this.Values.Where( p => !p.ExtId.HasValue ).ToList(); }
    }

    /// <summary>
    /// Liste der externen Projekte
    /// </summary>
    public IList<Project> ExternalProjects
    {
      get { return this.Values.Where( p => p.ExtId.HasValue ).ToList(); }
    }

    #endregion

    #region Public Methods

    /// <summary>
    /// Fügt ein Projekt in den Container ein
    /// </summary>
    /// <param name="project">Projekt das eingefügt wird</param>
    public void Add( Project project )
    {
      this.Add( project.Id, project );
    }

    public void Add( Projects projects )
    {
      foreach ( Project project in projects.Values )
        this.Add( project.Id, project );
    }

    public void Remove( Projects projects )
    {
      foreach ( Project project in projects.Values )
        this.Remove( project.Id );
    }

    /// <summary>
    /// Führt die Continer zusammen, indem alle übergebenen Projekte in diese Collection aufgenommen werden.
    /// Ist ein Projekt in beiden Container enthalten, wird die Projektinstanz dieses Containers mit den
    /// Projektdaten des übergebenen aktualisiert.
    /// </summary>
    /// <param name="projects">Projekte die in diese Collection zusammengeführt werden.</param>
    public void Merge( Projects projects )
    {
      foreach ( Project project in projects.Values ) {

        if ( this.ContainsKey( project.Id ) ) 
          this[project.Id].Assign( project );        

        else
          this.Add( project );
      }
    }

    public IList<Project> ToList()
    {
      return new List<Project>( this.Values );

    }

    #endregion

  }


}
